#############################################################################
# Author: Tom Pollard <pollard@schrodinger.com>
#
# Contributors: Dan Herron     <herron@cogsci.ucsd.edu>
#               David Schooley <schooley@ee.gatech.edu>
#               Vince Darley   <darley@fas.harvard.edu>
#               Martijn Koster <m.koster@nexor.co.uk>
#
# Version History:
# 3.4  1/98  -  Tom Pllard went over code and fixed things broken by 7.0, 
#                   completions tutorial added, 7.1b compatiblity made.
#               New 'marking proc's, and code sectioning support (i.e. dividers),
#                   plus reorganized code. -trf
# 3.12 10/97 -  Uses new menu-building code, so you can add with menu::insert (v)
# 3.11 9/97  -  Fixed problem with modevars in new Alpha scheme (Johan)
# 3.10 8/97  -  Modernised for new Alpha Tcl scheme (vince)
# 3.0  4/97  -  MacPerl interactions don't depend on MacPerl app name anymore
#               Fixed bug with perlFileAsFilter ($scriptStart uninitialized)
# 2.9  3/97  -  Fixed bug in command-dbl-click help lookup for Perl5 mode
# 2.8  2/97  -  Added Quick-Save commands in new submenu [Dan Herron]
#               "Save As CGI" finally works.
# 2.7  2/97  -  Comments before "#!/bin/perl" no longer confuse 'gotoPerlError'
# 2.6  2/97  -  Added electricPerlLeft and electricPerlRight - [David Schooley]
# 2.51 1/96  -  Fixed problem w/ "Tell MacPerl:Save As..."
# 2.5  1/96  -  Colorization and cmd-dbl-click modified to support Perl 5 docs
# 2.41 7/95  -  Minor tweaks
# 2.4  7/95  -  Fixed bugs affecting running unsaved scripts and error handling
# 2.3  7/95  -  Minor tweaks and code rearrangement.
# 2.2  6/95  -  Text filters act only on current line if "Apply to Buffer" is
#                  false and no text has been selected.
#               Bug fix in error-marking for scripts sent as AppleEvent params.
#               Cmd-dbl-clicking a function call jumps to function, if
#                  defined in the same file.
# 2.1  6/95  -  Cmd-dbl-clicking a 'require'd filename opens the file.
# 2.0  6/95  -  Minor bug fixes (incl. keyword decapitalization)
#               Alpha 6.0b17 compatibility updates.
#               Text Filters folder is settable from the App Paths menu now.
# 1.9  5/95  -  Cmd-dbl-clicking Perl keywords and special variables displays
#                  the man page info.
# 1.81 4/95  -  one very minor Alpha compatibility update (winInfo->getWinInfo).
# 1.8  4/95  -  Menu reorganized somewhat.
#               Text Filters folder can now be anywhere.
#               "ApplyToBuffer" flag ignored if text has been selected.
#               Bug fixes.
# 1.7  1/95  -  Updated to take advantage of MacPerl 4.1.4 AppleEvent features:
#                1) Text filters use 'batch' doScript (.: STDOUT file obsolete)
#                2) Filter scripts sent as doScript params (.: SCRIPT file obsolete)
#                3) "Save As Droplet" and "Save as Runtime" commands added.
#               Errors generated in 'require'd files are now displayed correctly
# 1.6 10/94  -  "UseDebugger" flag added (forces scripts to run under debugger).
#               Key bindings added for some menu commands.
#               "perlDoScript{,2,3}" procs consolidated into a single proc.
#               "saveAndRun" option added.
#               Command-line args now parsed into units more correctly, in
#                   particular, quoted file names aren't broken up.
#               "Close Output Window" added to "Tell MacPerl" menu.
#               Updated for Alpha 5.98 to load when menu is inserted.
#               The error messages window is now recycled.
#               "perlRecycleOutput" recycles output window.
#               Minor bug fixes.
# 1.5  9/94  -  MacPerl menu rearranged somewhat.
#               Explicit "Get Output Window" command added to menu.
#               Reading "#!" line for args is incompatible w/ standard,
#                   so it's been dropped.
#               Only scan the first 40 output lines for error messages (faster)
#				"wrapFilterScript" no longer opens STDIN
#               Text filters may now use command-line args
#               STDIN for text filters passed as explicit cmd-line arg 
# 1.4  9/94  -  The "#!" line of every script is read for command-line args,
#					which are passed explicitly to MacPerl with the script.
#				"PromptForArgs" menu flag added.
#				"perlCmdlineArgs" modeVar holds default command-line args.
#				Scripts are sent using custom "perlDoScript2" proc, which
#					allows passing of explicit command-line args.
# 1.3  9/94  -  When any script generates a compilation error, the file 
#					containing the script is brought up with the offending 
#					line highlighted; all error output is also written to
#					a "Perl Error Messages" window.
#				'repeatLastFilter' runs again the last text-filter script used.
#				'perlLastFilter' modeVar holds pathname of last filter.
#				Menu flags now mirrored as modeVars, so they can be saved and
#					restored between sessions.
#				Minor bug fixes.
# 1.2  8/94  -  'retrieveOutput' and 'autoSwitch' flags added.
#				'openInMacperl' added.
#				MacPerl output window now closed before new scripts are sent.
#				Filters now abort if there are compilation errors, and
#				MacPerl diagnostic output retrieved and displayed in Alpha.
# 1.1  8/94  -  'quitMacperl' added.
#               perl-mode file-marking updated for Alpha 5.90
#               Simplified installation via 'loadMacperl'(Pete Keleher). 
# 1.0  7/94  -  perl-mode setup updated for Alpha 5.85:
#                    keyword colorization supported
#                    custom file-marking added
#               #! lines in filter scripts now handled correctly 
#               Workarounds installed for AppleEvent bug in MacPerl 4.1.3
# 0.9  3/94  -  perl-mode stuff added, and
#               highlighted 'Perl commands' file (man page) prepared
#               minor bug fixes, too
# 0.8  3/94  -  flags are now check-marked
# 0.7  3/94  -  nested Text Filters folder now supported
#               menu format modified somewhat
# 0.6  3/94  -  'applyToBuffer' flag added
#               scripts in Alpha buffers can now be used as filters 
# 0.5  2/94  -  'filters', 'open special' submenu added
#               'overwrite' flag added
# 0.2  1/94  -  menu support added (Martijn Koster <m.koster@nexor.co.uk>)
#               'execute selection', 'execute buffer' commands added
# 0.1  9/93  -  text filter functionality created
#                  
##############################################################################
