#############################################################################
#############################################################################
#
# latex.tcl, version 4.6:  macros, menu, and key bindings for LaTeX users
#
#############################################################################
#
# version 1.1 and 1.2 (11/10/92) by Richard T. Austin <austin@eecs.umich.edu>
# versions 2.0--3.2 and 3.2t (3/97) by Tom Scavo <trscavo@syr.edu>
# versions 4.0 onwards (9/97) by Vince Darley <darley@fas.harvard.edu>
# If you make improvements to this file, please share them!
#
# LaTeX 2.09 is no longer supported.  If someone wants to make the 
# necessary changes to support it, they are welcome.
#############################################################################
#############################################################################

# Initialize TeX mode:
alpha::mode TeX 4.6.6 texMenu {
    *.tex *.ltx *.dtx *.ins *.cls *.clo *.sty *.log *.aux *.bbl *.idx 
    *.ind *.glo *.gls *.toc *.lof *.lot *.blg *.ilg *.fd *.drv
} { 
    texMenu electricReturn electricTab
} { 
    addMenu texMenu "270" TeX
    set unixMode(latex) {TeX}
    
    alpha::package require -loose Alpha 7.1
    # Reset the search path so it will be rebuilt the next time it's needed:
    set AllTeXSearchPaths {}
    
    # Initialize TeX-related application signatures:
    foreach app {tex viewDVI printDVI dvips viewPS printPS bibtex makeindex} {
	ensureset ${app}Sig {}
    }
    
    # Install the short LaTeX menu, by default:
    ensureset useShortLaTeXMenu 1
    
    # Do not load AMS-LaTeX, by default:
    ensureset useAMSLaTeX 0
    
    # To add a new fileset type, all we have to do is this:
    set fileSetsTypesThing(tex) "tex"
    lappend fileSetsTypes "tex"
} maintainer {
    "Vince Darley" darley@fas.harvard.edu <http://www.fas.harvard.edu/~darley/>
} uninstall {this-directory} help {file "LaTeX Help"}

#############################################################################
#
# Autoload procedures
#
#############################################################################

proc dummyTeX {} {}
proc texMenu {} {}

# Update this preference:
if {[info exists TeXmodeVars(TeXInputsFolder)]} {
    alertnote "The 'TeXInputsFolder' pref has been replaced by\
      'TeXSearchPath', which is now a list of folders to search.\
      I'll copy over your preference and delete the old one."
    lappend TeXmodeVars(TeXSearchPath) $TeXmodeVars(TeXInputsFolder)
    unset TeXmodeVars(TeXInputsFolder)
    lappend modifiedArrayElements [list TeXInputsFolder TeXmodeVars] \
      [list TeXSearchPath TeXmodeVars]
}

#############################################################################
#
# Flags and variables
#
#############################################################################

# Flags ---

# If a particular template requires a certain package, warn if it's not there.
newPref f warnIfPackageNotIncluded 1 TeX
# Build the submenu of all known packages.  Can be slow.
newPref f buildPkgsSubmenu 0 TeX
newPref f deleteEnvNoisily 1 TeX
newPref f deleteObjNoisily 1 TeX
# When latex mode asks you for something in the status bar, it will 'beep'
# to catch your attention.
newPref f promptNoisily 1 TeX
# Typeset occurs in background by default (you can always override with
# a different key-combination
newPref f runTeXInBack 0 TeX makeTypesetMenu
# When your actions require the mode to search for something, 'beep'
# if the search fails (otherwise just post a silent message).
newPref f searchNoisily 1 TeX
# Turn '...' into '\ldots' automatically as you type
newPref f smartDots 1 TeX
# Turn ' or " into ` or `` or ' or '' automatically as you type, depending
# upon the context
newPref f smartQuotes 1 TeX
newPref f useBrackets 0 TeX
# Use '$...$' and '$$...$$' for mathematics rather than the LaTeX
# expressions
newPref f useDollarSigns 0 TeX shadowUseDollarSigns
# Prompt in the status bar if possible (rather than use a dialog)
newPref f useStatusBar 1 TeX
newPref f wordWrap 1 TeX

# names of commands which insert boxed graphics
newPref v boxMacroNames {includegraphics includegraphics*} TeX shadowBoxMacroNames
# commands used by your citation package
newPref v citeCommands {cite nocite citet citeauthor citep citeyear} TeX shadowCiteRefCommands
# the list of known mathematical environments
newPref v mathEnvironments [list "math" "displaymath" "equation" "eqnarray" "eqnarray*" "array"] TeX
newPref v fillColumn {70} TeX
newPref v funcExpr {\\(sub)*section(\[.*\]|\*)?({[^{}]*})?} TeX
newPref v parseExpr {[^\{]*\{(.*)\}[^\}]*} TeX
newPref v funcExprAlt {\\section(\[.*\]|\*)?({[^{}]*})?} TeX 
newPref v leftFillColumn {0} TeX
newPref v prefixString {% } TeX
# the commands which are used to refer to labels elsewhere in the text
newPref v refCommands {ref eqref pageref vref vpageref} TeX shadowCiteRefCommands
newPref v wrapBreak {(\w+|\\(\\\*?|[^A-Za-z\t\r*\s\]|[A-Za-z]+\*?))} TeX
newPref v wrapBreakPreface {([^\w\]|.\\)} TeX
newPref v wordBreak {(([\w:-])+|\\(\\|[a-zA-Z]+)\*?|\\[^a-zA-Z*\s])} TeX
if {[info tclversion] < 8.0} {
    newPref v wordBreakPreface "(\[^-\\w:\\\]|.\\\\)" TeX
} else {
    newPref v wordBreakPreface "(\[^-\\w:\\\]|.\\\\|^|\$)" TeX
}
set texParaCommands {\[|\]|begin|end|(protect\\)?label|(sub)*section\*?|subfigure|paragraph|centerline|centering|caption|chapter|item|bibitem|intertext|(protect\\)?newline|includegraphics\*?}
# Where your TeX application searches for input files
newPref v TeXSearchPath "" TeX shadowTeXInputs
# The delimiter to use by default in labels inserted into the text.
# For instance the default is a colon, so labels are written 
# \label{fig:...} or \label{eq:...}.  You may prefer a hyphen '-' or...
newPref v standardTeXLabelDelimiter ":" TeX \
  {hook::callAll TeX::labelDelimChanged *}

# Pressing this key will insert a subscript _{} if the context is relevant
newPref binding smartSubscripts "/_<U" TeX "" 1
# Pressing this key will insert a superscript _{} if the context is relevant
newPref binding smartSuperscripts "/^<U" TeX "" 1
# If possible, show TeX log inside Alpha (doesn't work with all TeX apps)
newPref variable showTeXLog 1 TeX "" \
  [list "Never" "Only after error" "Always"] index
# Chooses between using the Marks menu for a list of TeX commands or
# the structural content of the file.
newPref variable markCommandsNotStructure \
  [list .sty .dtx .cls .clo .def .cfg .fd] TeX
# Don't indent lines between \begin,\end pairs. This only works for
# automatic indentation after you press return or tab, not for blocks
# of code inserted on the fly by code templates for example.
newPref flag dontIndentBeginEnd 0 TeX

#############################################################################
#
# Initialization
#
#############################################################################

hook::register activateHook makeTypesetMenu TeX
hook::register saveasHook makeTypesetMenu TeX
proc makeTypesetMenu {args} {
    menu::buildSome Typeset
}

# registered only if we're using Textures
proc TexturescloseHook {name} {
    global Texturesconnections
    if {[info exists Texturesconnections]} {
	set winNames [map car $Texturesconnections]
	if { [set pos [lsearch -exact $winNames $name]] > -1 } {
	    # Close Textures' connection:
	    global texAppSignatures
	    set TeXjob [cadr [lindex $Texturesconnections $pos]]
	    AEBuild -r '$texAppSignatures(Textures)' BSRs Disc Jobi $TeXjob
	    # Update list of connections:
	    set Texturesconnections [lreplace $Texturesconnections $pos $pos]
	}
    }
}

proc TexturesLaunched {args} {
    global Texturesconnections
    set Texturesconnections ""	
}

# The proc 'shadowtexSig' is executed automatically whenever 
# the global variable 'texSig' is altered, say, by choosing
# 'removedSavedSetting' on the Config menu.  Alpha now has
# a nifty proc to handle all the hard work for us: (Vince)
proc shadowtexSig {args} {
    global packagesSubmenuItems AllTeXSearchPaths texSig texAppSignatures
    # Reset the Packages submenu so that it will be rebuilt:
    set packagesSubmenuItems ""
    message {Building LaTeX menu}
    buildLaTeXMenuQuietly
    # Reset the search path so it will be rebuilt the next time it's needed:
    set AllTeXSearchPaths {}
    # Set things up for Textures
    if {[info exists texAppSignatures(Textures)] && \
      $texSig == $texAppSignatures(Textures)} {
        hook::register launch TexturesLaunched $texAppSignatures(Textures)
        hook::register closeHook TexturescloseHook TeX
    } else {
        hook::deregister launch TexturesLaunched $texAppSignatures(Textures)
        hook::deregister closeHook TexturescloseHook TeX
    }
    
}

proc shadowTeXInputs {args} {
    global packagesSubmenuItems AllTeXSearchPaths
    
    # Reset the Packages submenu so that it will be rebuilt:
    set packagesSubmenuItems ""
    
    message "Rebuilding the Documents submenu"
    menu::buildSome Documents
    message ""
    
    # Reset the search path so it will be rebuilt the next time it's needed:
    set AllTeXSearchPaths {}
}

# The proc 'shadowUseDollarSigns' is executed automatically whenever 
# the flag 'useDollarSigns' is altered.  The proc 'mathModeSubmenu'
# is defined in latexMenu.tcl.
proc shadowUseDollarSigns {args} {
    TeX::sub::MathModes
    menu::buildSome "Math Modes"
}

# we indent to previous non-comment line, +- a bit for \begin or \end
proc TeX::indentLine {} {
    set beg [lineStart [getPos]]
    set text [getText $beg [nextLineStart $beg]]
    regexp "^\[ \t\]*" $text white
    set next [pos::math $beg + [string length $white]]
    set lwhite [TeX::correctIndentation [getPos] \
      [getText $next [pos::math $next + 5]]]
    
    set lwhite [text::indentOf $lwhite]
    if {$white != $lwhite} {
	replaceText $beg $next $lwhite
    }
    goto [pos::math $beg + [string length $lwhite]]
}

proc TeX::correctIndentation {pos {next ""}} {
    global indentationAmount commentsArentSpecialWhenIndenting dontIndentBeginEnd
    set beg [lineStart $pos]

    # find last non-empty, non-comment line
    if {$commentsArentSpecialWhenIndenting} {
	if {[catch {search -s -f 0 -r 1 -i 0 -m 0 -- "^\[ \t\]*\[^ \t\n\r\]" [pos::math $beg - 1]} lst]} {
	    return 0
	}
    } else {
	if {[catch {search -s -f 0 -r 1 -i 0 -m 0 -- "^\[ \t\]*\[^ \t\n\r%\]" [pos::math $beg - 1]} lst]} {
	    return 0
	}
    }
    
    set line [getText [lindex $lst 0] [pos::math [nextLineStart [lindex $lst 0]] - 1]]
    set lwhite [posX [pos::math [lindex $lst 1] - 1]]	
    if {!$dontIndentBeginEnd} {
	incr lwhite [expr {$indentationAmount * \
	  ([regexp {^[ \t]*\\begin\{} $line] - [regexp {^\\end\{} $next])}]
    }
    # only happens with poorly formatted files.
    if {$lwhite < 0} {return 0} else {return $lwhite}
}

#############################################################################
#
# Colorization
#
#############################################################################

newPref v keywordColor blue TeX
newPref v sectionColor magenta TeX
newPref v commentColor red TeX
newPref v bracesColor green TeX

# The proc 'shadowCiteRefCommands' is executed whenever the variables 
# TeXmodeVars(citeCommands) or TeXmodeVars(refCommands) are changed.
proc shadowCiteRefCommands {args} {
    global TeXmodeVars
    set LaTeXClickWords {}
    # insure that 'citeCommands' contains "cite" and "nocite":
    lunion TeXmodeVars(citeCommands) cite nocite
    foreach word $TeXmodeVars(citeCommands) { 
	lappend LaTeXClickWords "\\\\$word" 
    }
    # insure that 'refCommands' contains "ref" and "pageref":
    lunion TeXmodeVars(refCommands) ref pageref
    foreach word $TeXmodeVars(refCommands) { 
	lappend LaTeXClickWords "\\\\$word" 
    }
    # color and underline command-double-clickable LaTeX commands:
    regModeKeywords -a -u -k $TeXmodeVars(keywordColor) TeX $LaTeXClickWords
}

# The proc 'shadowBoxMacroNames' is executed whenever the variable 
# TeXmodeVars(boxMacroNames) is changed.
proc shadowBoxMacroNames {args} {
    global TeXmodeVars
    set LaTeXClickWords {}
    foreach word $TeXmodeVars(boxMacroNames) { 
	lappend LaTeXClickWords "\\\\$word" 
    }
    # color and underline command-double-clickable LaTeX commands:
    regModeKeywords -a -u -k $TeXmodeVars(keywordColor) TeX $LaTeXClickWords
}

proc colorLaTeXClickCommands {} {
    global TeXmodeVars
    lappend LaTeXClickWords \\documentclass \\usepackage \
      \\input \\include \\InputIfFileExists \
      \\bibliography \\bibliographystyle \
      \\LoadClass \\RequirePackage
    # color and underline command-double-clickable LaTeX commands:
    regModeKeywords -a -u -k $TeXmodeVars(keywordColor) TeX $LaTeXClickWords
    shadowCiteRefCommands
    shadowBoxMacroNames
}

proc colorLaTeXCommands {} {
    global TeXmodeVars 
    set LaTeXKeyWords {}
    # Color comments, braces and dollar signs, and all LaTeX commands:
    regModeKeywords -i "\}" -i "\{" -i "\$" -I $TeXmodeVars(bracesColor) \
      -e {%} -m {\\} -c $TeXmodeVars(commentColor) \
      -k $TeXmodeVars(keywordColor) TeX $LaTeXKeyWords
    # Color sectioning commands:
    regModeKeywords -a -k $TeXmodeVars(sectionColor) TeX \
      { \\part \\chapter \\section \\subsection \\subsubsection \\paragraph }
}

colorLaTeXCommands; rename colorLaTeXCommands ""
colorLaTeXClickCommands; rename colorLaTeXClickCommands ""

#############################################################################
#
# LaTeX initialisation
#
#############################################################################

if {[alpha::tryToLoad "Initialising LaTeX" \
  latexEngine.tcl {} \
  latexComm.tcl {}\
  latexMacros.tcl {}\
  latexMenu.tcl {} \
  latexKeys.tcl {} \
  bindLaTeXKeys {Binding LaTeX keys} \
  [list app::registerMultiple tex [list OTEX *TEX *XeT MPS*] \
  [list 266 267 272 299] shadowtexSig] {Building LaTeX menu}]} {
    rename bindLaTeXKeys ""
}












